﻿/*
 * project : AcknexWrapper 1.0.2 for V7.82.3
 * created by : Matthias Auer
 * contact : mail@matthias-auer.net
 * homepage : http://acknexwrapper.matthias-auer.net/
 * date : 05.04.10
 */

using System;
using System.Runtime.InteropServices;
using System.Collections;
using System.Collections.Generic;
using AcknexWrapper.Native;

namespace AcknexWrapper
{
    /// <summary>
    /// used by : on_client, SchedulerForLevelEditorEntity, ent_create
    /// </summary>
    public delegate IEnumerable EventIntPtr(IntPtr p);

    /// <summary>
    /// used by : on_anykey, on_joy, on_level, on_mouse_left, on_mouse_right, on_mouse_middle
    /// </summary>
    public delegate IEnumerable EventVar(Var p);

    /// <summary>
    /// used by : on_click, on_close, on_d3d_.., on_exit, on_f1...on_f12, on_a...on_z, on_0...on_9, on_load, on_mouse_stop, pan_setevent, pan_setbutton, MATERIAL-event, on_maximize, on_minimize, on_resize
    /// </summary>
    public delegate IEnumerable EventVoid();

    /// <summary>
    /// used by : on_server
    /// </summary>
    public delegate IEnumerable EventOnServer(IntPtr str, Var id);

    /// <summary>
    /// used by : on_message
    /// </summary>
    public delegate int EventOnMessage(IntPtr hwnd, uint message, IntPtr WParam, IntPtr LParam);

    /// <summary>
    /// used by : effect, effect_local, PARTICLE-event
    /// </summary>
    public delegate void EventPARTICLE(PARTICLE p);

    namespace Native
    {
        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        internal delegate void NativeEventIntPtr(IntPtr p);

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        internal delegate void NativeEventVar(NativeVar p);

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        internal delegate void NativeEventVoid();

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        internal delegate void NativeEventOnServer(IntPtr str, NativeVar id);

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        internal delegate int NativeEventOnMessage(IntPtr hwnd, uint message, IntPtr WParam, IntPtr LParam);

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        internal delegate void NativeEventPARTICLE(IntPtr p);
    }

    public static class Scheduler
    {
        public enum PROC_MODE
        {
            /// <summary>
            /// Move to the list that is executed as first
            /// </summary>
            PROC_EARLIEST,
            /// <summary>
            /// Move to the list that is executed as second
            /// </summary>
            PROC_EARLY,
            /// <summary>
            /// Move to the list that is executed as third
            /// </summary>
            PROC_NORMAL,
            /// <summary>
            /// Move to the list that is executed as fourth
            /// </summary>
            PROC_LATE,
            /// <summary>
            /// Move to the list that is executed as fifth
            /// </summary>
            PROC_LATEST
        }

        private static Object lockObject = new Object();

        /// <summary>
        /// enable Scheduler
        /// </summary>
        public static Boolean SchedulerEnabled = true;

        /// <summary>
        /// stores all functions that will be scheduled
        /// </summary>
        private static LinkedList<ScheduleListElement> EarliestScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> EarlyScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> NormalScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> LateScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> LatestScheduleList = new LinkedList<ScheduleListElement>();

        /// <summary>
        /// stores all functions that will be added to the ScheduleLists
        /// </summary>
        private static LinkedList<ScheduleListElement> EarliestAddToScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> EarlyAddToScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> NormalAddToScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> LateAddToScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> LatestAddToScheduleList = new LinkedList<ScheduleListElement>();

        /// <summary>
        /// stores all functions that will be deleted from the ScheduleLists
        /// </summary>
        private static LinkedList<ScheduleListElement> EarliestDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> EarlyDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> NormalDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> LateDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
        private static LinkedList<ScheduleListElement> LatestDeleteFromScheduleList = new LinkedList<ScheduleListElement>();

        /* references */
        private static LinkedList<EventEngObjIntPtr> KeepEngObjReferenceList = new LinkedList<EventEngObjIntPtr>();
        private static LinkedList<EventEngObjIntPtr> AddToKeepEngObjReferenceList = new LinkedList<EventEngObjIntPtr>();
        private static LinkedList<EventEngObjIntPtr> DeleteFromKeepEngObjReferenceList = new LinkedList<EventEngObjIntPtr>();

        /* particle */
        private static LinkedList<EventParticleReference> KeepParticleReferenceList = new LinkedList<EventParticleReference>();
        private static LinkedList<EventParticleReference> AddToKeepParticleReferenceList = new LinkedList<EventParticleReference>();
        private static LinkedList<EventParticleReference> DeleteFromKeepParticleReferenceList = new LinkedList<EventParticleReference>();

        /* add functions thread safe */
        private static LinkedList<EventVoid> AddFunctionsThreadSafeList = new LinkedList<EventVoid>();

        internal static bool EngineClosed = false;

        private static EventVoidReference MainFunctionReference;

        private static IEnumerator ScheduleListEnumerator;
        private static IEnumerator AddToScheduleListEnumerator;
        private static IEnumerator DeleteFromScheduleListEnumerator;
        private static ScheduleListElement CurrentListElement;

        private static IEnumerator AddToKeepEngObjReferenceListEnumerator;
        private static IEnumerator DeleteFromKeepEngObjReferenceListEnumerator;
        private static IEnumerator KeepParticleReferenceListEnumerator;
        private static IEnumerator AddToKeepParticleReferenceListEnumerator;
        private static IEnumerator DeleteFromKeepParticleReferenceListEnumerator;

        /// <summary>
        /// reset all scheduler lists
        /// </summary>
        internal static void ResetScheduler()
        {
            EarliestScheduleList = new LinkedList<ScheduleListElement>();
            EarlyScheduleList = new LinkedList<ScheduleListElement>();
            NormalScheduleList = new LinkedList<ScheduleListElement>();
            LateScheduleList = new LinkedList<ScheduleListElement>();
            LatestScheduleList = new LinkedList<ScheduleListElement>();

            EarliestAddToScheduleList = new LinkedList<ScheduleListElement>();
            EarlyAddToScheduleList = new LinkedList<ScheduleListElement>();
            NormalAddToScheduleList = new LinkedList<ScheduleListElement>();
            LateAddToScheduleList = new LinkedList<ScheduleListElement>();
            LatestAddToScheduleList = new LinkedList<ScheduleListElement>();

            EarliestDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
            EarlyDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
            NormalDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
            LateDeleteFromScheduleList = new LinkedList<ScheduleListElement>();
            LatestDeleteFromScheduleList = new LinkedList<ScheduleListElement>();


            KeepEngObjReferenceList = new LinkedList<EventEngObjIntPtr>();
            AddToKeepEngObjReferenceList = new LinkedList<EventEngObjIntPtr>();
            DeleteFromKeepEngObjReferenceList = new LinkedList<EventEngObjIntPtr>();
            KeepParticleReferenceList = new LinkedList<EventParticleReference>();
            AddToKeepParticleReferenceList = new LinkedList<EventParticleReference>();
            DeleteFromKeepParticleReferenceList = new LinkedList<EventParticleReference>();
        }
        

        /// <summary>
        /// This element contains the IEnuemrator method that has to be called, an the time to the next call (NextSchedule)
        /// </summary>
        private class ScheduleListElement
        {
            public IEnumerator ElementToSchedule;
            public float NextSchedule;

            public ScheduleListElement(IEnumerator ElementToSchedule, float NextSchedule)
            {
                this.ElementToSchedule = ElementToSchedule;
                this.NextSchedule = NextSchedule;
            }
        }
        
        /// <summary>
        /// this method tries to cast an object to a float value
        /// </summary>
        /// <param name="ObjectToCast">ObjectToCast the object to cast</param>
        /// <returns>the casted float</returns>
        private static float ObjectToFloat(Object ObjectToCast, LinkedList<ScheduleListElement> ListToDeleteFrom, ScheduleListElement ListElement)
        {
            if (ListToDeleteFrom == null)
            {
                NormalAddToScheduleList.AddFirst(ListElement);
            }
            if (ObjectToCast.GetType() == typeof(Var))
            {
                return (Var)ObjectToCast;
            }
            else
                if (ObjectToCast.GetType() == typeof(int))
                {
                    return (int)ObjectToCast;
                }
                else
                    if (ObjectToCast.GetType() == typeof(float))
                    {
                        return (float)ObjectToCast;
                    }
                    else
                        if (ObjectToCast.GetType() == typeof(PROC_MODE))
                        {
                            if (ListToDeleteFrom != null)   //if the element was already in a scheduler list
                            {
                                ListToDeleteFrom.AddFirst(ListElement); //delete from current list
                            }
                            else //this case happens when the method is added to the scheduler but not yet in a list and yield return returns a PROC_MODE
                            {
                                NormalAddToScheduleList.Remove(ListElement);
                            }
                            switch ((PROC_MODE)ObjectToCast)
                            {
                                case PROC_MODE.PROC_EARLIEST:
                                    EarliestAddToScheduleList.AddFirst(ListElement);
                                    break;
                                case PROC_MODE.PROC_EARLY:
                                    EarlyAddToScheduleList.AddFirst(ListElement);
                                    break;
                                case PROC_MODE.PROC_NORMAL:
                                    NormalAddToScheduleList.AddFirst(ListElement);
                                    break;
                                case PROC_MODE.PROC_LATE:
                                    LateAddToScheduleList.AddFirst(ListElement);
                                    break;
                                case PROC_MODE.PROC_LATEST:
                                    LatestAddToScheduleList.AddFirst(ListElement);
                                    break;
                            }
                            return 1;   //wait one frame
                        }
                        else
                        {
                            new Exception("The yield return value has to be of type Var, int, float or PROC_MODE");
                            return 0;
                        }
        }

        /// <summary>
        /// special class that allows to schedule a function that is not called by the engine
        /// </summary>
        private class CustomAddedFunction : ScheduleListElement
        {
            private EventVoid EventVoidReferenceStorage;

            public CustomAddedFunction(EventVoid MethodToSchedule) : base(MethodToSchedule().GetEnumerator(), 0)
            {
                EventVoidReferenceStorage = MethodToSchedule;
                if (ElementToSchedule.MoveNext() == true) //if there's a "wait" (yield) in the method, add the method to the SchedulerList
                {
                    this.NextSchedule = ObjectToFloat(ElementToSchedule.Current, null, this);   //store the time until the next schedule is due
                }
            }
        }

        /// <summary>
        /// Add the method passed as parameter to the scheduler
        /// </summary>
        public static void AddEventVoid(EventVoid MethodToSchedule)
        {
            new CustomAddedFunction(MethodToSchedule);
        }

        /// <summary>
        /// Use this method if you want to use the 'AddEventVoid' method from another thread - the method is maybe called first, in the next frame !
        /// </summary>
        /// <param name="MethodToSchedule">the method to add</param>
        public static void AddEventVoidThreadSafe(EventVoid MethodToSchedule)
        {
            lock (lockObject)
            {
                AddFunctionsThreadSafeList.AddLast(MethodToSchedule);
            }
        }

        #region EventDefinitions

        /// <summary>
        /// delete an Event storage that points to entity that an Event is assigned
        /// </summary>
        private static void DeleteAlreadyExistingReference(IntPtr EngObjIntPtr)
        {
            IEnumerator traverseList = AddToKeepEngObjReferenceList.GetEnumerator();
            while (traverseList.MoveNext() == true)
            {
                if (((EventEngObjIntPtr)traverseList.Current).EngObjIntPtr == EngObjIntPtr)
                {
                    DeleteFromKeepEngObjReferenceList.AddFirst((EventEngObjIntPtr)traverseList.Current);
                }
            }
        }

        internal class EventEngObjIntPtr
        {
            public IntPtr EngObjIntPtr = IntPtr.Zero;
        }

        /// <summary>
        /// used to keep a reference to a delegate that otherwise would be deleted by the garbage collector
        /// </summary>
        internal class EventVoidReference : EventEngObjIntPtr
        {
            private EventVoid EventVoidReferenceStorage;
            public NativeEventVoid NativeEventVoidFunc;

            /// <summary>
            /// standard constructor for storing an Event and providing an Event
            /// </summary>
            public EventVoidReference(EventVoid EventVoidReference)
            {
                this.EventVoidReferenceStorage = EventVoidReference;
                this.NativeEventVoidFunc = NativeEventVoidFuncOriginal;
            }

            /// <summary>
            /// constructor for a EngObj
            /// </summary>
            public EventVoidReference(EventVoid EventVoidReference, IntPtr EngObjIntPtr)
            {
                this.EventVoidReferenceStorage = EventVoidReference;
                this.EngObjIntPtr = EngObjIntPtr;
                this.NativeEventVoidFunc = NativeEventVoidFuncOriginal;
                AddToKeepEngObjReferenceList.AddFirst(this);    //if this event is used by an EngObj we need to keep the reference in a list
            }

            /// <summary>
            /// the message that is called by the engine
            /// </summary>
            private void NativeEventVoidFuncOriginal()
            {
                if (SchedulerEnabled == true)
                {
                    IEnumerator EventEnumerator = EventVoidReferenceStorage().GetEnumerator();
                    if (EventEnumerator.MoveNext() == true) //if there's a "wait" (yield) in the method, add the method to the SchedulerList
                    {
                        ScheduleListElement TempElement = new ScheduleListElement(EventEnumerator, 1);
                        TempElement.NextSchedule = ObjectToFloat(EventEnumerator.Current, null, TempElement);
                    }
                }
                else
                {
                    EventVoidReferenceStorage().GetEnumerator().MoveNext();   //if a scheduler isn't used - only call the method
                }
            }
        }

        /// <summary>
        /// used to keep a reference to a delegate that otherwise would be deleted by the garbage collector
        /// </summary>
        internal class EventVarReference : EventEngObjIntPtr
        {
            private EventVar EventVarReferenceStorage;
            public NativeEventVar NativeEventVarFunc;

            /// <summary>
            /// standard constructor for storing an Event and providing an Event
            /// </summary>
            public EventVarReference(EventVar EventVarReference)
            {
                this.EventVarReferenceStorage = EventVarReference;
                this.NativeEventVarFunc = NativeEventVarFuncOriginal;
            }

            /// <summary>
            /// constructor for a EngObj
            /// </summary>
            public EventVarReference(EventVar EventVarReference, IntPtr EngObjIntPtr)
            {
                this.EventVarReferenceStorage = EventVarReference;
                this.EngObjIntPtr = EngObjIntPtr;
                this.NativeEventVarFunc = NativeEventVarFuncOriginal;
                AddToKeepEngObjReferenceList.AddFirst(this);    //if this event is used by an EngObj we need to keep the reference in a list
            }

            /// <summary>
            /// the message that is called by the engine
            /// </summary>
            private void NativeEventVarFuncOriginal(NativeVar p)
            {
                if (SchedulerEnabled == true)
                {
                    IEnumerator EventEnumerator = EventVarReferenceStorage((Var)p).GetEnumerator();
                    if (EventEnumerator.MoveNext() == true) //if there's a "wait" (yield) in the method, add the method to the SchedulerList
                    {
                        ScheduleListElement TempElement = new ScheduleListElement(EventEnumerator, 1);
                        TempElement.NextSchedule = ObjectToFloat(EventEnumerator.Current, null, TempElement);
                    }
                }
                else
                {
                    EventVarReferenceStorage((Var)p).GetEnumerator().MoveNext();   //if a scheduler isn't used - only call the method
                }
            }
        }

        /// <summary>
        /// Use this function if you want to add a function to an entity that was created in the level editor
        /// </summary>
        public static void SchedulerForLevelEditorEntity(EventIntPtr EventIntPtrReference, ENTITY Entity)
        {
            if (Entity == null)
            {
                throw new ArgumentNullException("parameter");
            }
            EventIntPtrReference EntityEventReference = new EventIntPtrReference(EventIntPtrReference, (IntPtr)Entity);
            EntityEventReference.NativeEventIntPtrFunc((IntPtr)Entity); //start the function as if the engine had called it
        }

        /// <summary>
        /// used to keep a reference to a delegate that otherwise would be deleted by the garbage collector
        /// </summary>
        internal class EventIntPtrReference : EventEngObjIntPtr
        {
            private EventIntPtr EventIntPtrReferenceStorage;
            public NativeEventIntPtr NativeEventIntPtrFunc;

            /// <summary>
            /// standard constructor for storing an Event and providing an Event
            /// </summary>
            public EventIntPtrReference(EventIntPtr EventIntPtrReference)
            {
                this.EventIntPtrReferenceStorage = EventIntPtrReference;
                this.NativeEventIntPtrFunc = NativeEventIntPtrFuncOriginal;
            }

            /// <summary>
            /// constructor for a EngObj
            /// </summary>
            public EventIntPtrReference(EventIntPtr EventIntPtrReference, IntPtr EngObjIntPtr)
            {
                this.EventIntPtrReferenceStorage = EventIntPtrReference;
                this.EngObjIntPtr = EngObjIntPtr;
                this.NativeEventIntPtrFunc = NativeEventIntPtrFuncOriginal;
                AddToKeepEngObjReferenceList.AddFirst(this);    //if this event is used by an EngObj we need to keep the reference in a list
            }

            /// <summary>
            /// constructor for a EngObj that is a PARTICLE
            /// </summary>
            public EventIntPtrReference(IntPtr EngObjIntPtr, EventIntPtr EventIntPtrReference)
            {
                this.EventIntPtrReferenceStorage = EventIntPtrReference;
                this.EngObjIntPtr = EngObjIntPtr;
                this.NativeEventIntPtrFunc = NativeEventIntPtrFuncOriginal;
                AddToKeepEngObjReferenceList.AddFirst(this);    //if this event is used by an EngObj we need to keep the reference in a list
            }

            /// <summary>
            /// the message that is called by the engine
            /// </summary>
            private void NativeEventIntPtrFuncOriginal(IntPtr p)
            {
                if (SchedulerEnabled == true)
                {
                    IEnumerator EventEnumerator = EventIntPtrReferenceStorage(p).GetEnumerator();
                    if (EventEnumerator.MoveNext() == true) //if there's a "wait" (yield) in the method, add the method to the SchedulerList
                    {
                        ScheduleListElement TempElement = new ScheduleListElement(EventEnumerator, 1);
                        TempElement.NextSchedule = ObjectToFloat(EventEnumerator.Current, null, TempElement);
                    }
                }
                else
                {
                    EventIntPtrReferenceStorage(p).GetEnumerator().MoveNext(); ;   //if a scheduler isn't used - only call the method
                }
            }
        }

        /// <summary>
        /// used to keep a reference to a delegate that otherwise would be deleted by the garbage collector
        /// </summary>
        internal class EventOnServerReference : EventEngObjIntPtr
        {
            private EventOnServer EventOnServerReferenceStorage;
            public NativeEventOnServer NativeEventOnServerFunc;

            /// <summary>
            /// standard constructor for storing an Event and providing an Event
            /// </summary>
            public EventOnServerReference(EventOnServer EventOnServerReference)
            {
                this.EventOnServerReferenceStorage = EventOnServerReference;
                this.NativeEventOnServerFunc = NativeEventOnServerFuncOriginal;
            }

            /// <summary>
            /// the message that is called by the engine
            /// </summary>
            private void NativeEventOnServerFuncOriginal(IntPtr p, NativeVar id)
            {
                if (SchedulerEnabled == true)
                {
                    IEnumerator EventEnumerator = EventOnServerReferenceStorage(p, (Var)id).GetEnumerator();
                    if (EventEnumerator.MoveNext() == true) //if there's a "wait" (yield) in the method, add the method to the SchedulerList
                    {
                        ScheduleListElement TempElement = new ScheduleListElement(EventEnumerator, 1);
                        TempElement.NextSchedule = ObjectToFloat(EventEnumerator.Current, null, TempElement);
                    }
                }
                else
                {
                    EventOnServerReferenceStorage(p, (Var)id).GetEnumerator().MoveNext();   //if a scheduler isn't used - only call the method
                }
            }
        }

        /// <summary>
        /// used to keep a reference to a delegate that otherwise would be deleted by the garbage collector
        /// </summary>
        internal class EventOnMessageReference : EventEngObjIntPtr
        {
            private EventOnMessage EventOnMessageReferenceStorage;
            public NativeEventOnMessage NativeEventOnMessageFunc;

            /// <summary>
            /// standard constructor for storing an Event and providing an Event
            /// </summary>
            public EventOnMessageReference(EventOnMessage EventOnMessageReference)
            {
                this.EventOnMessageReferenceStorage = EventOnMessageReference;
                this.NativeEventOnMessageFunc = NativeEventOnMessageFuncOriginal;
            }

            /// <summary>
            /// the message that is called by the engine
            /// </summary>
            private int NativeEventOnMessageFuncOriginal(IntPtr hwnd, uint message, IntPtr WParam, IntPtr LParam)
            {
                return EventOnMessageReferenceStorage(hwnd, message, WParam, LParam);
            }
        }

        /// <summary>
        /// used to keep a reference to a delegate that otherwise would be deleted by the garbage collector - for PARTICLE
        /// </summary>
        internal class EventParticleReference
        {
            private EventPARTICLE EventParticleReferenceStorage;
            public NativeEventPARTICLE NativeEventParticleFunc;
            public bool UsedInThisCycle = true;

            /// <summary>
            /// standard constructor for storing an Event and providing an Event
            /// </summary>
            public EventParticleReference(EventPARTICLE EventIntPtrReference)
            {
                this.EventParticleReferenceStorage = EventIntPtrReference;
                this.NativeEventParticleFunc = NativeEventParticleFuncOriginal;
                AddToKeepParticleReferenceList.AddFirst(this);
            }

            /// <summary>
            /// the message that is called by the engine
            /// </summary>
            private void NativeEventParticleFuncOriginal(IntPtr p)
            {
                UsedInThisCycle = true; //alive if still called
                EventParticleReferenceStorage((PARTICLE)p);
            }
        }

        #endregion

        internal static void RemoveFromKeepReferenceList(IntPtr p)
        {
            IEnumerator traverseList = KeepEngObjReferenceList.GetEnumerator();
            while (traverseList.MoveNext() == true) //traverse the List and delete all objects that reference to the same EngObj
            {
                if (((EventEngObjIntPtr)traverseList.Current).EngObjIntPtr == p)
                {
                    DeleteFromKeepEngObjReferenceList.AddFirst((EventEngObjIntPtr)traverseList.Current);
                }
            }
        }

        private static void ScheduleList(LinkedList<ScheduleListElement> ListToSchedule, LinkedList<ScheduleListElement> ListToDeleteFrom)
        {
            //the functions that are to be scheduled
            ScheduleListEnumerator = ListToSchedule.GetEnumerator();  //get an Enumerator for the ScheduleList
            while (ScheduleListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                CurrentListElement = (ScheduleListElement)ScheduleListEnumerator.Current;
                if (CurrentListElement.NextSchedule > 0)
                {
                    CurrentListElement.NextSchedule -= 1;   //reduce time
                    CurrentListElement.NextSchedule = Math.Max(CurrentListElement.NextSchedule, 0);
                }
                else
                {
                    CurrentListElement.NextSchedule += EngVar.time_step.FloatValue / 16;
                    CurrentListElement.NextSchedule = Math.Min(CurrentListElement.NextSchedule, 0);
                }
                if (CurrentListElement.NextSchedule == 0)
                {
                    if (CurrentListElement.ElementToSchedule.MoveNext() == true) //schedule this element
                    {
                        CurrentListElement.NextSchedule = ObjectToFloat(CurrentListElement.ElementToSchedule.Current, ListToDeleteFrom, CurrentListElement);
                    }
                    else //if the method ended remove this function from the schedule list
                    {
                        ListToDeleteFrom.AddFirst(CurrentListElement);
                    }
                }
            }
        }

        private static void AddToScheduleList(LinkedList<ScheduleListElement> ScheduleList, LinkedList<ScheduleListElement> AddToScheduleList)
        {
            AddToScheduleListEnumerator = AddToScheduleList.GetEnumerator();    //get an Enumerator for the AddToScheduleList
            while (AddToScheduleListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                if (ScheduleList == EarlyScheduleList || ScheduleList == EarliestScheduleList)
                {
                    ScheduleList.AddFirst((ScheduleListElement)AddToScheduleListEnumerator.Current);  //add the element at the start of the list
                }
                else
                {
                    ScheduleList.AddLast((ScheduleListElement)AddToScheduleListEnumerator.Current);  //add the element at the end of the list
                }
            }
            AddToScheduleList.Clear();  //we may now clear this list
        }

        private static void DeleteFromList(LinkedList<ScheduleListElement> ScheduleList, LinkedList<ScheduleListElement> DeleteFromScheduleList)
        {
            DeleteFromScheduleListEnumerator = DeleteFromScheduleList.GetEnumerator();    //get an Enumerator for the DeleteFromScheduleList
            while (DeleteFromScheduleListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                ScheduleList.Remove((ScheduleListElement)DeleteFromScheduleListEnumerator.Current);  //remove the element from the list
            }
            DeleteFromScheduleList.Clear();  //we may now clear this list
        }

        /// <summary>
        /// starts the Scheduler and the Method that is given as Parameter (if one is given)
        /// </summary>
        public static void StartScheduler(EventVoid MainMethod)
        {
            /* start the MainMethod in the Scheduler */
            if (MainMethod != null)
            {
                MainFunctionReference = new EventVoidReference(MainMethod);
                MainFunctionReference.NativeEventVoidFunc();
            }

            while ((EngineClosed == false) && (EngFun.engine_frame() != 0))
            {
                ScheduleList(EarliestScheduleList, EarliestDeleteFromScheduleList);
                ScheduleList(EarlyScheduleList, EarlyDeleteFromScheduleList);
                ScheduleList(NormalScheduleList, NormalDeleteFromScheduleList);
                ScheduleList(LateScheduleList, LateDeleteFromScheduleList);
                ScheduleList(LatestScheduleList, LatestDeleteFromScheduleList);

                AddToScheduleList(EarliestScheduleList, EarliestAddToScheduleList);
                AddToScheduleList(EarlyScheduleList, EarlyAddToScheduleList);
                AddToScheduleList(NormalScheduleList, NormalAddToScheduleList);

                lock (lockObject)
                {
                    foreach (EventVoid eventVoid in AddFunctionsThreadSafeList)
                    {
                        AddEventVoid(eventVoid);
                    }
                    AddFunctionsThreadSafeList.Clear();
                }

                AddToScheduleList(LateScheduleList, LateAddToScheduleList);
                AddToScheduleList(LatestScheduleList, LatestAddToScheduleList);

                DeleteFromList(EarliestScheduleList, EarliestDeleteFromScheduleList);
                DeleteFromList(EarlyScheduleList, EarlyDeleteFromScheduleList);
                DeleteFromList(NormalScheduleList, NormalDeleteFromScheduleList);
                DeleteFromList(LateScheduleList, LateDeleteFromScheduleList);
                DeleteFromList(LatestScheduleList, LatestDeleteFromScheduleList);

                CleanUpReferences();
            }
            EngVar.ResetReferences();   //set all EngVar references to null
            Scheduler.ResetScheduler(); //set all Scheduler lists to new lists, thus empty
            NativeEngFun.engine_close();
        }

        /// <summary>
        /// This method cleans up unused "Delegate" objects, it is called every cycle by the scheduler.
        /// </summary>
        public static void CleanUpReferences()
        {
            /* EventReferenceStorage */

            AddToKeepEngObjReferenceListEnumerator = AddToKeepEngObjReferenceList.GetEnumerator();    //get an Enumerator for the AddToScheduleList
            while (AddToKeepEngObjReferenceListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                KeepEngObjReferenceList.AddLast((EventEngObjIntPtr)AddToKeepEngObjReferenceListEnumerator.Current);  //add the element at the end of the list
            }
            AddToKeepEngObjReferenceList.Clear();  //we may now clear this list

            DeleteFromKeepEngObjReferenceListEnumerator = DeleteFromKeepEngObjReferenceList.GetEnumerator();    //get an Enumerator for the DeleteFromScheduleList
            while (DeleteFromKeepEngObjReferenceListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                KeepEngObjReferenceList.Remove((EventEngObjIntPtr)DeleteFromKeepEngObjReferenceListEnumerator.Current);  //remove the element from the list
            }
            DeleteFromKeepEngObjReferenceList.Clear();  //we may now clear this list

            /* Particle */

            KeepParticleReferenceListEnumerator = KeepParticleReferenceList.GetEnumerator();    //get an Enumerator for the AddToScheduleList
            while (KeepParticleReferenceListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                if (((EventParticleReference)KeepParticleReferenceListEnumerator.Current).UsedInThisCycle == false)
                {
                    DeleteFromKeepParticleReferenceList.AddLast((EventParticleReference)KeepParticleReferenceListEnumerator.Current);  //add to the delete list if it wasn't used in this cycle
                }
                ((EventParticleReference)KeepParticleReferenceListEnumerator.Current).UsedInThisCycle = false;
            }

            AddToKeepParticleReferenceListEnumerator = AddToKeepParticleReferenceList.GetEnumerator();    //get an Enumerator for the AddToScheduleList
            while (AddToKeepParticleReferenceListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                KeepParticleReferenceList.AddLast((EventParticleReference)AddToKeepParticleReferenceListEnumerator.Current);  //add the element at the end of the list
            }
            AddToKeepParticleReferenceList.Clear();  //we may now clear this list

            DeleteFromKeepParticleReferenceListEnumerator = DeleteFromKeepParticleReferenceList.GetEnumerator();    //get an Enumerator for the DeleteFromScheduleList
            while (DeleteFromKeepParticleReferenceListEnumerator.MoveNext() == true)   //as long as we haven't yet iterated over all elements of the list
            {
                KeepParticleReferenceList.Remove((EventParticleReference)DeleteFromKeepParticleReferenceListEnumerator.Current);  //remove the element from the list
            }
            DeleteFromKeepParticleReferenceList.Clear();  //we may now clear this list
        }
    }
}